# -*- coding: utf-8 -*-

from sendmail import send_mail

def enviar_email(remitente, destinatarios, asunto, mensaje, 
                 servidor, login, password, ficheros=[], html='', cc=[], bcc=[], 
                 charset='iso-8859-1'):
    
    send_mail(remitente, destinatarios, asunto, mensaje, servidor, login, password,
              files=ficheros, html=html, cc=cc, bcc=bcc, charset=charset)
    
def normalizar_a_html(texto):
    return texto.\
        replace('\'', '&#39;').\
        replace('&', '&amp;').\
        replace('#', '&#35;').\
        replace('á', '&aacute;').\
        replace('Á', '&Aacute;').\
        replace('é', '&eacute;').\
        replace('É', '&Eacute;').\
        replace('í', '&iacute;').\
        replace('Í', '&Iacute;').\
        replace('ó', '&oacute;').\
        replace('Ó', '&Oacute;').\
        replace('ú', '&uacute;').\
        replace('Ú', '&Uacute;').\
        replace('ü', '&uuml;').\
        replace('Ü', '&Uuml;').\
        replace('ñ', '&ntilde;').\
        replace('Ñ', '&Ntilde;').\
        replace('<', '&lt;').\
        replace('>', '&gt;').\
        replace('¡', '&iexcl;').\
        replace('?', '&iquest;').\
        replace('"', '&quot;').\
        replace('%', '&#37;')
        


# -*- coding: utf-8 -*-

IMPLTYPE_POSTGRESQL = 'postgres'
IMPLTYPE_POSTGRESQL2 = 'postgres2'
IMPLTYPE_FIREBIRD = 'firebird'

CONF_HOST = 'host'
CONF_DB = 'db'
CONF_USER = 'user'
CONF_PASSW = 'password'

# sexos
SEXO_HOMBRE = 1
SEXO_MUJER = 2

# dias
LUNES = 1
MARTES = 2
MIERCOLES = 3
JUEVES = 4
VIERNES = 5
SABADO = 6
DOMINGO = 7

# privilegios
PRIVL_NINGUNO = 'NINGUNO'
PRIVL_LECTURA = 'READ-ONLY'
PRIVL_LECTURA_ESCRITURA = 'READ-WRITE'

# claves definidas para la definición de privilegios
CAMPOS = 'campos'
ETIQUETA = 'etiqueta'
ORDEN = 'orden'
UTILIZAR_SELECTOR = 'utilizar_selector'
BUSQUEDA = 'busqueda'
COD_CLASE = 'cod_clase'
ORDENACION = 'ordenacion'
CLASE = 'cls'

NOMBRE = 'nombre'
TABLA_RELACIONADA = 'tabla_relacionada'
TIPO = 'tipo'
VALOR = 'valor'
VALORES = 'valores'
SOLO_LECTURA = 'solo_lectura'
REQUERIDO = 'requerido'
TITULO = 'titulo'
CAMPOS_REFERENCIA = 'campos_referencia'

TYPE_DATE = 'date'
TYPE_TIME = 'time'
TYPE_BOOLEAN = 'boolean'
TYPE_STRING = 'string'
TYPE_CHAR = 'char'
TYPE_TEXT = 'text'
TYPE_INTEGER = 'integer'
TYPE_REAL = 'real'

# constantes para Firebird
VAR_LANZADOR = 'lanzador'
VAR_TIMEOUT_CONSULTAS = 'time_out_consultas'

#
STATUS = 'status'
MESSAGE = 'message'
EXPIRATION = 'expiration'

# idiomas
LANG_ES_ES = 0
LANG_EN_UK = 1



# -*- coding: utf-8 -*-

from neptuno.dataset import DataSet
from sqlalchemy import Table, MetaData, select, exists
from sqlalchemy.sql.expression import and_, Select, or_, alias
from sqlalchemy.types import DATE, TIME, DATETIME, INTEGER, NUMERIC, BOOLEAN, \
    BIGINT
import re
import datetime as dt

_CONSTANT_DATE_TODAY       = 'today'
_CONSTANT_DATE_START_WEEK  = 'start_week'
_CONSTANT_DATE_END_WEEK    = 'end_week'
_CONSTANT_DATE_START_MONTH = 'start_month'
_CONSTANT_DATE_END_MONTH   = 'end_month'
_CONSTANT_DATE_START_YEAR  = 'start_year'
_CONSTANT_DATE_END_YEAR    = 'end_year'
_CONSTANT_DAY              = 'd'
_CONSTANT_WEEK             = 'w'
_CONSTANT_MONTH            = 'm'
_CONSTANT_YEAR             = 'y'

class Busqueda(object):
    
    def __init__(self, tabla, texto_busqueda, columnas_trans=None, strtodatef=None):
        self.tabla = tabla
        self.cols = [col.name for col in self.tabla.columns]
        self.cache_campo = {}
        
        # string to date conversion function
        self.strtodatef = strtodatef
        
        if not columnas_trans:
            self.cols_trans = [re.sub(r'[^a-z0-9]*', '', self.quitar_acentos(col))
                               for col in self.cols]
        else:
            self.cols_trans = columnas_trans
            
        if texto_busqueda:
            
            if isinstance(texto_busqueda, str):
                texto_busqueda = texto_busqueda.decode('utf-8').lower().encode('utf-8')
                
            elif isinstance(texto_busqueda, unicode):
                texto_busqueda = texto_busqueda.lower().encode('utf-8')
                
            self.condicion = self.condicion_busqueda(texto_busqueda)
            self.orden = self.orden_busqueda(texto_busqueda)
            
        else:
            self.condicion = self.orden = None

    def sin_acentos(self, texto):
        
        resultado = texto
        resultado = re.sub(r'(a|á|à|ä)', '(a|á|à|ä)', resultado)
        resultado = re.sub(r'(e|é|è|ë)', '(e|é|è|ë)', resultado)
        resultado = re.sub(r'(i|í|ì|ï)', '(i|í|ì|ï)', resultado)
        resultado = re.sub(r'(o|ó|ò|ö)', '(o|ó|ò|ö)', resultado)
        resultado = re.sub(r'(u|ú|ù|ü)', '(u|ú|ù|ü)', resultado)
        
        return resultado
        
    def quitar_especiales(self, texto):
        texto = texto.encode('utf-8')
        
        return texto.\
            replace("'", "''").\
            replace('(', ' ').replace(')', ' ').\
            replace('?', ' ').replace('¿', ' ').\
            replace('@', ' ')
            
    def quitar_acentos(self, texto):
        texto = texto.encode('utf-8')
        
        return texto.lower().\
            replace('á', 'a').\
            replace('é', 'e').\
            replace('í', 'i').\
            replace('ó', 'o').\
            replace('ú', 'u').\
            replace('ü', 'u')
            
    def orden_busqueda(self, busqueda):
        terminos = busqueda.decode('utf-8').split(',')

        orden = []
        for termino in terminos:
            termino = termino.strip()
            if termino != '':
                m  = re.search(r'^(\+|-)(\w+)', termino)                
                if m is None:
                    continue
                
                operador = m.group(1)
                
                campos = termino.split(operador)
                
                c = self.get_col(campos[1].strip())
                if c:
                    if operador == '+':
                        operador = ' ASC'
                    elif operador == '-':
                        operador = ' DESC'
                        
                    orden.append('"%s" %s' % (c, operador))
                        
#        print 'ORDEN =' + '\n'.join(orden)
        return ', '.join(orden)
    
    def get_col(self, campo):
        
        if self.cache_campo.has_key(campo.lower()):
            return self.cache_campo[campo.lower()]
        
        else:
            col = None
            for ct, c in zip(self.cols_trans, self.cols):
                
                if ct.startswith(campo.lower()):
                    if not col or col['length'] > len(ct):
                        col = dict(name=c, length=len(ct))
                        
            if col:
                self.cache_campo[campo.lower()] = col['name']
                return col['name']
            
            else:
                return None
            
    def process_date_constants(self, termino):
        
        def _sub(m):
            cons = m.group(1)
            s = m.group(3)
            q = int(m.group(4) or 0)
            t = m.group(5) or _CONSTANT_DAY
            
            r = None
            if cons == _CONSTANT_DATE_TODAY:
                r = dt.date.today()
            
            elif cons == _CONSTANT_DATE_START_WEEK:
                wd = dt.date.today().weekday()
                r = dt.date.today() - dt.timedelta(days=wd)
            
            elif cons == _CONSTANT_DATE_END_WEEK:
                wd = 6 - dt.date.today().weekday()
                r = dt.date.today() + dt.timedelta(days=wd)
            
            elif cons == _CONSTANT_DATE_START_MONTH:
                today = dt.date.today()
                r = dt.date(today.year, today.month, 1)
            
            elif cons == _CONSTANT_DATE_END_MONTH:
                today = dt.date.today()
                next_month = today + dt.timedelta(days=30)
                r = dt.date(next_month.year, next_month.month, 1) - dt.timedelta(days=1)
            
            elif cons == _CONSTANT_DATE_START_YEAR:
                today = dt.date.today()
                r = dt.date(today.year, 1, 1)
            
            elif cons == _CONSTANT_DATE_END_YEAR:
                today = dt.date.today()
                r = dt.date(today.year, 12, 31)
            
            if r:    
                if s and q:
                    if t == _CONSTANT_DAY:
                        if s == '-':
                            n = -q
                        else:
                            n = q
                        
                        r = r + dt.timedelta(days=n)
                            
                    elif t == _CONSTANT_WEEK:
                        if s == '-':
                            n = -q*7
                        else:
                            n = q*7
                        
                        r = r + dt.timedelta(days=n)
                    
                    elif t == _CONSTANT_MONTH:
                        
                        r_ = r
                        
                        if s == '-':
                            n = -1
                        else:
                            n = +1
                            
                        i = 0
                        while abs(i) != q:
                            next_month = r_ + dt.timedelta(days=30*n)
                            r_ = dt.date(next_month.year, next_month.month, r.day)
                            
                            i += n
                            
                        r = r_
                            
                    elif t == _CONSTANT_YEAR:
                        if s == '-':
                            n = -q
                            
                        else:
                            n = +q
                            
                        r = dt.date(r.year + n, r.month, r.day)
                        
                return r.strftime('%Y-%m-%d')
            
            else:
                return m.group(0)
        
        return re.sub(r'\{\s*(\w+)(\s+([\+\-])\s*(\d+)\s*(\w{0,1}))?\s*\}', _sub, termino)
    
    def busqueda_campos(self, busqueda):
        
        terminos = self.process_date_constants(busqueda).split(',')
        
        resultado = None
        sql = []
        for termino in terminos:
            termino = termino.strip()
            if termino != '':
                
                m = re.search(r'^(\w*)\s*(<>|<=|>=|==|=|<|>|#|!)(.+)', termino)                
                if m is None:
                    continue

                operador = m.group(2)
                campos = termino.split(operador)
                
                if operador == '#' or operador == '!':
                    campo = m.group(3).strip()
                    
                else:
                    campo = m.group(1).strip()
                    if m.group(3).strip() == '':
                        continue
                    
                c = self.get_col(campo)
                if not c:
                    continue
                
                col = self.tabla.columns[c]
                
                if operador in ['#', '!']:
                    
                    # fecha y hora
                    if isinstance(col.type, DATE) or \
                    isinstance(col.type, TIME) or \
                    isinstance(col.type, DATETIME):
                        # no es nulo
                        if operador == '#':
                            # "Fecha inicio" IS NOT NULL
                            sql.append('"%s" IS NOT NULL' % c.encode('utf-8'))
                            
                        # es nulo
                        elif operador == '!':
                            # "Fecha inicio" IS NULL
                            sql.append('"%s" IS NULL' % c.encode('utf-8'))

                        elif operador == '==':
                            sql.append('"%s" = \'%s\'' % (c.encode('utf-8'),
                                                          campos[1].strip()))

                    # números (integer, biginteger, numeric)
                    elif isinstance(col.type, INTEGER) or \
                    isinstance(col.type, NUMERIC) or \
                    isinstance(col.type, BIGINT):
                    
                        # no es nulo
                        if operador == '#':
                            # "Nº de alumnos" IS NOT NULL
                            sql.append('"%s" IS NOT NULL' % c.encode('utf-8'))
                            
                        # es nulo
                        elif operador == '!':
                            # "Nº de alumnos" IS NULL
                            sql.append('"%s" IS NULL' % c.encode('utf-8'))
                            
                        elif operador == '==':
                            # Nº de alumnos = 10
                            sql.append('"%s" = %s' % (c.encode('utf-8'),
                                                      campos[1].strip()))
                            
                    # booleano
                    elif isinstance(col.type, BOOLEAN):
                        
                        if operador == '#':
                            sql.append('"%s" = True' % c.encode('utf-8'))
                            
                        elif operador == '!':
                            sql.append('"%s" = False' % c.encode('utf-8'))                                         

                    # texto (o cualquier otra cosa)
                    else:
                        # no es nulo
                        if operador == '#':
                            # trim(coalesce("---" as Text), '')) <> ''
                            sql.append('TRIM(COALESCE(CAST("%s" as Text), \'\')) <> \'\'' % \
                                       c.encode('utf-8'))
                            
                        # es nulo
                        elif operador == '!':
                            # trim(coalesce("---" as Text), '')) = ''
                            sql.append('TRIM(COALESCE(CAST("%s" as Text), \'\')) = \'\'' % \
                                       c.encode('utf-8'))
                            
                        elif operador == '==':
                            sql.append('TRIM(COALESCE(CAST("%s" as Text), \'\')) = \'%s\'' % \
                                        (c.encode('utf-8'),
                                         campos[1]))                                         

                else:
                    explicitos = re.findall(r'"([^"]+)"', campos[1], re.I)
                    
                    # quitar los explícitos y los caracteres "especiales"
                    texto = self.quitar_especiales(re.sub(r'"[^"]+"', '', campos[1], re.I))
                    
                    terminos2 = texto.split()
                    expresiones = []
                    
                    expresion_novacio = 'TRIM(COALESCE(CAST("%s" as Text), \'\')) <> \'\'' % c.encode('utf-8')

                    for termino2 in terminos2:
                        termino2 = self.sin_acentos(termino2)
                        
                        # fecha, hora o fecha-hora
                        if isinstance(col.type, DATE) or \
                        isinstance(col.type, TIME) or \
                        isinstance(col.type, DATETIME):
                        
                            # convert to date
                            if self.strtodatef:
                                try:
                                    termino2 = self.strtodatef(termino2).strftime('%Y-%m-%d')
                                    #print termino2
                                    
                                except:
                                    pass
                        
                            if operador in ['=', '<>']:
                                expresion_novacio = None
                                
                                # "Fecha inicio" = '01/01/2010'                   
                                expresion = '"%s" %s \'%s\'' % \
                                    (c.encode('utf-8'),
                                     operador.encode('utf-8'),
                                     termino2)

                            else:
                                # <, >, <=, >=
                                expresion_novacio = '"%s" IS NOT NULL' % c.encode('utf-8')
                                
                                # "Fecha inicio" >= '01/01/2010'
                                expresion = '"%s" %s \'%s\'' % \
                                    (c.encode('utf-8'), 
                                     operador.encode('utf-8'), 
                                     termino2)
                                    
                        # número
                        elif isinstance(col.type, INTEGER) or \
                        isinstance(col.type, NUMERIC) or \
                        isinstance(col.type, BIGINT):

                            if operador in ['=', '<>']:
                                # "Nº de alumnos" = 10                   
                                expresion = '"%s" %s %s' % \
                                    (c.encode('utf-8'),
                                     operador.encode('utf-8'),
                                     termino2)

                            else:
                                # <, >, <=, >=
                                expresion_novacio = '"%s" IS NOT NULL' % c.encode('utf-8')

                                # "Nº de alumnos" >= 10
                                expresion = '"%s" %s %s' % \
                                    (c.encode('utf-8'), 
                                     operador.encode('utf-8'), 
                                     termino2)

                        # texto (o cualquier otra cosa)
                        else: 
                            if operador == '=':                            
                                expresion = "UPPER(CAST(\"%s\" as Text)) SIMILAR TO UPPER('%%%s%%')" % \
                                                (c.encode('utf-8'), termino2)

                            elif operador == '<>':
                                expresion = "UPPER(CAST(\"%s\" as Text)) NOT SIMILAR TO UPPER('%%%s%%')" % \
                                                (c.encode('utf-8'), termino2)

                            else:
                                # <, >, <=, >=
                                expresion_novacio = '"%s" IS NOT NULL' % c.encode('utf-8')

                                expresion = "\"%s\" %s '%s'" % \
                                    (c.encode('utf-8'), operador.encode('utf-8'), termino2)

                        expresiones.append(expresion)
                        
                    # explícitos ("---")
                    for exp in explicitos:
                        
                        if not (isinstance(col.type, DATE) or 
                                isinstance(col.type, TIME) or
                                isinstance(col.type, DATETIME)):

                            if operador == '=':
                                expresion = "UPPER(CAST(\"%s\" as Text)) LIKE UPPER('%%%s%%')" % \
                                                (c.encode('utf-8'), exp.encode('utf-8'))

                            elif operador == '<>':
                                expresion = "UPPER(CAST(\"%s\" as Text)) NOT LIKE UPPER('%%%s%%')" % \
                                                (c.encode('utf-8'), exp.encode('utf-8'))

                        else:
                            # date, time and datetime

                            # convert to date
                            if self.strtodatef:
                                try:
                                    termino2 = self.strtodatef(termino2).strftime('%Y-%m-%d')
                                    #print termino2

                                except:
                                    pass

                            if operador in ['=', '<>']:
                                expresion_novacio = None

                                # "Fecha inicio" = '01/01/2010'
                                expresion = '"%s" %s \'%s\'' % \
                                    (c.encode('utf-8'), 
                                     operador.encode('utf-8'), 
                                     exp.encode('utf-8'))

                            else:
                                # <, >, <=, >=
                                expresion_novacio = '"%s" IS NOT NULL' % c.encode('utf-8')

                                # "Fecha inicio" >= '01/01/2010'
                                expresion = '"%s" %s \'%s\'' % \
                                    (c.encode('utf-8'), 
                                     operador.encode('utf-8'), 
                                     exp.encode('utf-8'))

                        expresiones.append(expresion)
                    
                    expresion_novacio = ('%s AND \n(' % expresion_novacio 
                                         if expresion_novacio else '')
                    
                    if expresion_novacio:
                        sql.append('(' + expresion_novacio + ' AND\n'.join(expresiones) + '))')
                        
                    else:
                        sql.append('(' + ' AND\n'.join(expresiones) + ')')
                            
        if len(sql) > 0:
            resultado = ' AND\n'.join(sql).decode('utf-8')
        
        return resultado
                
    def condicion_busqueda(self, busqueda):
        
        if '*' in busqueda:
            sql = None
        
        else:            
            filtros = [f.replace("'", "''").strip() 
                       for f in busqueda.decode('utf-8').split(',')]
            
            sql = None
            sql_filtros = []
            for filtro in filtros:
                
                m = re.search(r'^([^"]*)(<>|<=|>=|==|=|<|>|#|!|\+|-)(.+)', filtro)  
                if m:
                    sql_filtro = self.busqueda_campos(filtro)
                    
                else:
                    sql_termino = []
                    
                    explicitos = re.findall(r'"([^"]+)"', filtro, re.I)
                    
                    filtro = re.sub(r'"[^"]+"', '', filtro, re.I | re.U)
                    
                    # quitar caracteres especiales
                    filtro = filtro.\
                        replace(u'(', u' ').replace(u')', u' ').\
                        replace(u'?', u' ').replace(u'¿', u' ').\
                        replace(u'@', u' ')
                    
                    terminos_de_busqueda = filtro.split()                    
                    
                    for termino in terminos_de_busqueda:
                                                
                        termino = self.sin_acentos(termino.encode('utf-8'))
                        
                        if termino.strip():
                            sql_campo = []
                            
                            for columna in self.cols:
                                if columna != 'id' and columna[0:3] != 'id_':
                                    try:
                                        texto = '(CAST("%s" as Text) <> \'\' AND ' % columna.encode('utf-8')
                                        texto += "UPPER(CAST(\"%s\" as Text)) SIMILAR TO UPPER('%%%s%%'))" % \
                                            (columna.encode('utf-8'), termino)
                                        
                                        sql_campo.append(texto)
                                    except:
                                        raise #return termino
                                
                            texto_sql_campo = ' OR\n'.join(sql_campo)
                            texto_sql_campo = ' (' + texto_sql_campo + ')'
                            
                            sql_termino.append(texto_sql_campo)
                    
                    # explícitos ("---")        
                    for exp in explicitos:
                        
                        termino = exp.encode('utf-8')
                        
                        if termino.strip():
                            sql_campo = []
                            
                            for columna in self.cols:
                                if columna != 'id' and columna != 'busqueda' and columna[0:3] != 'id_':
                                    try:
                                        texto = '(CAST("%s" as Text) <> \'\' AND ' % columna.encode('utf-8')
                                        texto += "UPPER(CAST(\"%s\" as Text)) LIKE UPPER('%%%s%%'))" % \
                                            (columna.encode('utf-8'), termino)
                                        
                                        sql_campo.append(texto)
                                    except:
                                        raise #return termino
                                
                            texto_sql_campo = ' OR\n'.join(sql_campo)
                            texto_sql_campo = ' (' + texto_sql_campo + ')'
                            
                            sql_termino.append(texto_sql_campo)
                            
                    sql_filtro = ' AND\n'.join(sql_termino)
                    sql_filtro = sql_filtro.decode('utf-8')
                    
                if sql_filtro:
                    sql_filtros.append(sql_filtro)
                    
            if sql_filtros:
                sql = ' AND\n'.join(sql_filtros)
        
        return sql

def search(session, table_name, q=None, rp=100, offset=0, show_ids=False, 
           strtodatef=None, filters=None, collection=None):
    """
    IN
      session     <sqlalchemy.orm.session.Session>
      table_name  <str>
      q           <str>
      rp          <int>
      offset      <int>
      show_ids    <bool> (opcional=False)
      strtodatef  <function> (opcional=None)
      filters     [<tuple>, ...]
      collecion   <tuple> (<str>, <str>, <int>,)
      
    OUT
      <DataSet>
    """ 
    
    meta = MetaData(bind=session.bind)
    
    if isinstance(table_name, Select):
        tbl = table_name
        
    else:
        tbl = Table(table_name, meta, autoload=True)
    
    sql = None
    order = ''
    if q:
        
        # process "q"
        qres = Busqueda(tbl, q, strtodatef=strtodatef)
        
        # apply search conditions
        sql = and_(qres.condicion, sql)
        
        # apply order
        if qres.orden:
            order = qres.orden
         
    # apply filters
    if filters:
        filters_tuple = (sql,)
        for f in filters:
            if len(f) > 2:
                # (<field name>, <field value>, <operator>,)
                # different
                if f[2] == '!=':
                    filters_tuple += (tbl.c[f[0]] != f[1],)
                    
                # greater
                elif f[2] == '>':
                    filters_tuple += (tbl.c[f[0]] > f[1],)
                
                # greater or equal
                elif f[2] == '>=':
                    filters_tuple += (tbl.c[f[0]] >= f[1],)
                    
                # less
                elif f[2] == '<':
                    filters_tuple += (tbl.c[f[0]] < f[1],)
                    
                # less or equal
                elif f[2] == '<=':
                    filters_tuple += (tbl.c[f[0]] <= f[1],)
                    
                # equal (and anything else...)
                else:
                    filters_tuple += (tbl.c[f[0]] == f[1],)
                    
            elif len(f) == 2:
                # (<field name>, <field value>,)
                # equal (by default) 
                filters_tuple += (tbl.c[f[0]] == f[1],)
                
            elif len(f) == 1:
                filters_tuple += (f[0],)
        
        sql = and_(*filters_tuple)
    
    if collection:
        
        child_table_name = collection[0]
        child_attr = collection[1]
        parent_id = collection[2]
        
        child_table = Table(child_table_name, meta, autoload=True)
        
        sel = select([child_table.c.id], from_obj=child_table,
                     whereclause=and_(child_table.c[child_attr] == parent_id,
                                      child_table.c.id == tbl.c.id,
                                      ),
                     ).correlate(tbl)
        
        sql = and_(sql, exists(sel))
            
    # where
    if isinstance(tbl, Select):
        qry = tbl.where(sql)
            
    else:
        qry = tbl.select(whereclause=sql)
        
    # order by
    if order:
        qry = qry.order_by(order)
        
    return DataSet.procesar_resultado(session, qry, rp, offset, show_ids)

class Search(object):
    
    def __init__(self, session, table_name, strtodatef=None):
        self.session = session
        self.table_name = table_name
        
        self.meta = MetaData(bind=self.session.bind)
        self.strtodatef = strtodatef
        self.sql = None
        self.order = ''
        
        if isinstance(self.table_name, Select):
            self.tbl = self.table_name
            
        else:
            self.tbl = Table(self.table_name, self.meta, autoload=True)
        
        self.from_ = self.tbl
            
    def apply_qry(self, q):
        
        if q:
            
            # process "q"
            qres = Busqueda(self.tbl, q, strtodatef=self.strtodatef)
            
            # apply search conditions    
            self.sql = and_(qres.condicion, self.sql)
            
            # apply order
            if qres.orden:
                self.order = qres.orden
                
    def and_(self, *cond):
        self.sql = and_(self.sql, *cond)
        
    def or_(self, cond):
        self.sql = or_(self.sql, cond)
        
    def join(self, *args, **kwargs):
        self.from_ = self.from_.join(*args, **kwargs)
        
    def outerjoin(self, *args, **kwargs):
        self.from_ = self.from_.outerjoin(*args, **kwargs)
                
    def apply_filters(self, filters):
        
        tbl = self.tbl
        
        # apply filters
        if filters:
            filters_tuple = (self.sql,)
            for f in filters:
                if len(f) > 2:
                    # (<field name>, <field value>, <operator>,)
                    # different
                    if f[2] == '!=':
                        filters_tuple += (tbl.c[f[0]] != f[1],)
                        
                    # greater
                    elif f[2] == '>':
                        filters_tuple += (tbl.c[f[0]] > f[1],)
                    
                    # greater or equal
                    elif f[2] == '>=':
                        filters_tuple += (tbl.c[f[0]] >= f[1],)
                        
                    # less
                    elif f[2] == '<':
                        filters_tuple += (tbl.c[f[0]] < f[1],)
                        
                    # less or equal
                    elif f[2] == '<=':
                        filters_tuple += (tbl.c[f[0]] <= f[1],)
                        
                    # equal (and anything else...)
                    else:
                        filters_tuple += (tbl.c[f[0]] == f[1],)
                        
                elif len(f) == 2:
                    # (<field name>, <field value>,)
                    # equal (by default) 
                    filters_tuple += (tbl.c[f[0]] == f[1],)
                    
                elif len(f) == 1:
                    filters_tuple += (f[0],)
            
            self.sql = and_(*filters_tuple)
        
    def __call__(self, rp=100, offset=0, collection=None, no_count=False, show_ids=False):
        """
        IN
          rp          <int>
          offset      <int>
          filters     [<tuple>, ...]
          collecion   <tuple> (<str>, <str>, <int>,)
          no_count    <bool> => False
          show_ids    <bool> => False
          
        OUT
          <DataSet>
        """ 
        
        sql = self.sql
        if collection:
            
            child_table_name = collection[0]
            child_attr = collection[1]
            parent_id = collection[2]
            
            child_table = alias(Table(child_table_name, self.meta, autoload=True))
            
            self.from_ = self.from_.\
                join(child_table,
                     and_(child_table.c.id == self.tbl.c.id,
                          child_table.c[child_attr] == parent_id))
            
        # where
        if isinstance(self.tbl, Select):
            qry = self.tbl.where(sql)
                
        else:
            qry = select([self.tbl], from_obj=self.from_, whereclause=sql)
            
        # order by
        if self.order:
            qry = qry.order_by(self.order)
            
        return DataSet.procesar_resultado(self.session, qry, rp, offset, 
                                          no_count=no_count, show_ids=show_ids)


# -*- coding: utf-8 -*-

from decimal import Decimal
from jinja2 import Environment, PackageLoader
from neptuno.ficherocsv import FicheroCSV
from neptuno.util import default_fmt_float
from neptuno.xlsreport import XLSReport
from sqlalchemy.types import INTEGER, NUMERIC, DATE, TIME, BOOLEAN, BIGINT, \
    DATETIME
import datetime as dt
import re
import sys

COLUMNAS = 'columnas'
COUNT = 'count'
LIMITE_RESULTADOS = 'limite_resultados'
DATOS = 'datos'
TOTALES = 'totales'

def tocsv(e):
    def _tocsv(func):
        def __tocsv(*args, **kwargs):
            return func(*args, **kwargs).to_csv(encoding=e)
        
        return __tocsv
    
    return _tocsv

def tostring(func):
    def __tostring(*args, **kwargs):
        return func(*args, **kwargs).tostring()
    
    return __tostring

def tojson(func):
    def __tojson(*args, **kwargs):
        return func(*args, **kwargs).to_json()
    
    return __tojson

def remove_specials(texto):
    return re.sub(r'[^a-zA-Z0-9_]', '_', texto)

class DataSetRowIterator(object):
    
    def __init__(self, row):
        self.row = row
        self.i = 0
        self.l = len(self.row.cols)
        
    def __iter__(self):
        return self
    
    def next(self):
        if self.i < self.l:
            item = self.row[self.i]
            self.i += 1
            
            return item
        else:
            raise StopIteration
    
    def __next__(self):
        return self.next()

class DataSetRow(object):
    
    cols = []
    
    def __init__(self, **kw):
        self.attr = kw
        
    def __str__(self):
        result = []
        for k, v in self.attr.iteritems():
            result.append('%s=%s' % (k, v))
            
        return ', '.join(result)
        
    def __getattr__(self, name):
        if self.attr.has_key(name):
            return self.attr[name]
        else:
            raise Exception('<DataSet>: member "%s" does not exist' % name)
        
    def __iter__(self):                
        return DataSetRowIterator(self)

    def __getitem__(self, key):
        if isinstance(key, int):
            k = self.cols[key]
            if sys.version_info < (2, 6, 6):
                if isinstance(k, unicode):
                    k = k.encode('utf-8')
            
            if self.attr.has_key(k):
                return self.attr[k]
            
            else:
                return None
        
        elif self.attr.has_key(key):
            return self.attr[key]
        
    def __setitem__(self, key, value):
        if isinstance(key, int):
            k = self.cols[key]
            if sys.version_info < (2, 6, 6):
                if isinstance(k, unicode):
                    k = k.encode('utf-8')
                            
            self.attr[k] = value
        
        else:
            if sys.version_info < (2, 6, 6):
                if isinstance(key, unicode):
                    key = key.encode('utf-8')
                    
            self.attr[key] = value
                
class DataSet(object):
    
    date_fmt = '%d/%m/%Y'
    time_fmt = '%H:%M'
    datetime_fmt = '%d/%m/%Y %H:%M:%S'
    true_const = True
    false_const = False
    int_fmt = '%d'
    
    def float_fmt(self, value):
        return default_fmt_float(value)
    
    def __init__(self, cols, types=None, limite=None, totales=None, **fmt):
        
        self.cols = []
        self.labels = []
        self.types = []
        
        # date
        if fmt.has_key('date_fmt'):
            self.date_fmt = fmt['date_fmt']
            
        # time
        if fmt.has_key('time_fmt'):
            self.time_fmt = fmt['time_fmt']
            
        # datetime
        if fmt.has_key('datetime_fmt'):
            self.datetime_fmt = fmt['datetime_fmt']
            
        # int
        if fmt.has_key('int_fmt'):
            self.int_fmt = fmt['int_fmt']
        
        # float
        if fmt.has_key('float_fmt'):
            self.float_fmt = fmt['float_fmt']
        
        for col in cols:
            self.append_col(col)

        self.limite_resultados = limite
        self.data = []
        self.count = 0
        self.totales = totales
        
    def append_col(self, col):
        if isinstance(col, tuple):
            # (<col>, <etiqueta>, <tipo>,) 
            self.cols.append(col[0])
            self.labels.append(col[1] or col[0])
            self.types.append(col[2])
            
        else:
            self.cols.append(col)
            self.labels.append(col)
            self.types.append('')

    def init_totales(self):
        if self.totales:
            totales = dict(zip(self.totales, [None] * len(self.totales)))
            
        else:
            totales = {}
            
        return totales
    
    def acumular(self, acumulado, valor):
        return (acumulado or 0) + (valor or 0)
        
    def append(self, dato=None):
        """
        IN
          dato
            <list>
            <dict>
            <DataSetRow>
            
        OUT
          ???
        """
        
        if isinstance(dato, DataSetRow):
            self.data.append(dato)
            
        elif isinstance(dato, dict):
            # compare keys and columns
            d = {}
            for k, v in dato.iteritems():
                if k not in self.cols:
                    raise Exception('The key "%s" is not correct' % k)
                
                if sys.version < (2, 6):
                    if isinstance(k, unicode):
                        k = k.encode('utf-8')
                        
                d[k] = v
            
            self.append(DataSetRow(**d))

        elif isinstance(dato, DataSet):
            for d in dato:
                self.append(d)
            
        elif isinstance(dato, list) or hasattr(dato, '__iter__'):
            d = {}
            for col, v in zip(self.cols, dato):
                if sys.version_info < (2, 6, 6):
                    if isinstance(col, unicode):
                        col = col.encode('utf-8')
                    
                d[col] = v
                
            self.append(DataSetRow(**d))

        else:
            self.append([None] * len(self.cols))        
        
        self.data[-1].cols = self.cols        
        return self.data[-1]
    
    def __iter__(self):
        return iter(self.data)
    
    def reversed(self):
        """Devuelve un iterador para recorrer la lista del final al principio
        
        Para un DataSet de N valores
        
        item0
        item1
        ...
        itemN-1
        
        devolver
        
        [
         itemN-1,
         ...
         item1,
         item0
        ]
        """
        return reversed(self.data)
    
    def ereversed(self):
        """
        Devuelve un iterador enumerado para recorrer la lista del final al 
        principio.
        
        Para un DataSet de N valores
        
        item0
        item1
        ...
        itemN-1
        
        devolver
        
        [
         (N-1, itemN-1)
         ...
         (1, item1)
         (0, item0)
        ]
        
        """
        return zip(range(len(self)-1, -1, -1), self.reversed())
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        return self.data[index]
    
    def __str__(self):
        return self.to_str()
    
    def getlabel(self, i):
        if not isinstance(i, int):
            i = self.cols.index(i)
        
        return self.label
    
    def gettype(self, i):
        if not isinstance(i, int):
            i = self.cols.index(i)
            
        return self.types[i]
    
    def _format(self, fmt, value):

        if isinstance(fmt, (str, unicode,)):
            s = unicode(fmt % value)
            
        elif hasattr(fmt, '__call__'):
            s = unicode(fmt(value))
            
        else:
            s = unicode(value or '')
            
        return s
    
    def _formatdt(self, fmt, value):

        if isinstance(fmt, (str, unicode,)):
            if isinstance(value, (dt.date, dt.datetime,)) and value >= dt.date(1900, 1, 1) or isinstance(value, dt.time):
                s = unicode(value.strftime(fmt))
                
            else:
                def _strftime(d, fmt):
                    return fmt.replace('%d', '%.2d' % d.day).\
                        replace('%m', '%.2d' % d.month).\
                        replace('%Y', '%.4d' % d.year)
                        
                s = unicode(_strftime(value, fmt))
            
        elif hasattr(fmt, '__call__'):
            s = unicode(fmt(value))
            
        else:
            s = unicode(value or '')
            
        return s

    def to_str(self, width=None, fit_width=True):
        """
        Devuelve los datos del DataSet en un cadena con formato tabular.
        
        |------|------|------|------|
        | (c1) | (c2) | (c3) | (c4) |
        |---------------------------|
        |      |      |      |      |
        |      |      |      |      |
        |      |      |      |      |
        |---------------------------|
        |      |      |      |      | <- totales
        |---------------------------|
        
        IN
          width <int> (opcional)
          Indica el ancho de las columnas
          
          fit_width <bool> (por defecto=True)
          Hace las columnas tan anchas como el valor m�s ancho, que no sobrepase
          el ancho fijado por 'width'.
        """
        
        if fit_width:
            widths = [0] * len(self.cols)
            
            for i, c in enumerate(self.cols):
                for d in self.data: 
                    if len(str(d[c])) > widths[i]:
                        widths[i] = len(unicode(d[c]))
                        
                if len(self.labels[i]) > widths[i]:
                    widths[i] = len(self.labels[i])

                if width and widths[i] > width:
                    widths[i] = width
        else:
            widths = [width or 10] * len(self.cols)
            
        resultado = ''
        
        cabecera = '|'
        for i, c in enumerate(self.labels):
            cabecera += unicode(c)[:widths[i]].center(widths[i]) + '|'
            
        l = len(cabecera)-2
            
        resultado += '|'+'-'*l + '|\n'
        resultado += cabecera + '\n'
        resultado += '|'+'-'*l + '|\n'
        
        totales = self.init_totales()
        
        for dato in self.data:
            linea = '|'
            for i, c in enumerate(self.cols):
                valor = dato[c]
                if valor is None:
                    valor = u''
                    
                w = widths[i]
                
                # totalizaciones
                if self.totales:
                    if totales.has_key(c):
                        totales[c] = self.acumular(totales[c], valor)
                
                # float/Decimal
                if isinstance(dato[c], (float, Decimal,)):
                    linea += self._format(self.float_fmt, dato[c])[:w].rjust(w)
                    
                # bool
                elif isinstance(dato[c], bool):
                    linea += unicode((self.true_const if dato[c] else self.false_const)).rjust(w)                   

                # int/long
                elif isinstance(dato[c], (int, long,)):
                    linea += self._format(self.int_fmt, dato[c])[:w].rjust(w)

                # datetime
                elif isinstance(dato[c], dt.datetime):
                    #linea += unicode(dato[c].strftime('%s %s' % (self.date_fmt, self.time_fmt)))[:w].rjust(w)
                    linea += unicode(self._formatdt(self.datetime_fmt, dato[c]))[:w].rjust(w)

                # date
                elif isinstance(dato[c], dt.date):
                    linea += unicode(self._formatdt(self.date_fmt, dato[c]))[:w].rjust(w)

                # time
                elif isinstance(dato[c], dt.time):
                    #linea += unicode(dato[c].strftime(self.time_fmt))[:w].rjust(w)
                    linea += unicode(self._formatdt(self.time_fmt, dato[c]))[:w].rjust(w)

                else:
                    linea += unicode(valor).replace('\n', '')[:w].ljust(w)
                     
                linea += '|'
            
            resultado += linea + '\n'
            
        resultado += '|'+'-'*l + '|\n'
        
        # línea de totales
        if self.totales:
            linea = '|'
            for i, c in enumerate(self.cols):
                w = widths[i]                
                if c in self.totales:                    
                    linea += self.float_fmt(totales[c])[:w].rjust(w)                    
                else:
                    linea += ''.rjust(w)
                    
                linea += '|'
                    
            resultado += linea + '\n'                    
            resultado += '|'+'-'*l + '|\n'
            
        return resultado
        
    def to_data(self, encoding=None):
        """
        """
        
        data = []
        
        for row in self.data:
            dato = []
            for item in row:
                
                # date
                if isinstance(item, dt.date):
                    dato.append(self._formatdt(self.date_fmt, item))
                    
                # time
                elif isinstance(item, dt.time):
                    dato.append(self._formatdt(self.time_fmt, item))
                
                # datetime
                elif isinstance(item, dt.datetime):
                    dato.append(self._formatdt(self.datetime_fmt, item))
                    
                # float, Decimal
                elif isinstance(item, (float, Decimal,)):
                    dato.append(self._format(self.float_fmt, item))
                    
                # bool
                elif isinstance(item, bool):
                    dato.append(self.true_const if item else self.false_const)
                    
                # int, long
                elif isinstance(item, int):
                    dato.append(self._format(self.int_fmt, item))
                
                # unicode
                elif isinstance(item, unicode):
                    dato.append(item)
                    
                # ???
                else:
                    dato.append(str(item or '').decode('utf-8'))
                    
            data.append(dato)
            
        return data
        
    def to_csv(self, encoding=None, mostrar_ids=False):
        """
        Devuelve en una cadena con formato CSV el contenido del DataSet.
        
        IN
          encoding <str> (opcional)
          
        OUT
          <str> (csv)
        """
        
        fichero_csv = FicheroCSV()
        fichero_csv.date_fmt = self.date_fmt
        fichero_csv.time_fmt = self.time_fmt
        fichero_csv.true_const = self.true_const
        fichero_csv.false_const = self.false_const
        fichero_csv.float_fmt = self.float_fmt
        
        columnas = []
        labels = []
        for item, c in enumerate(self.cols):
            
            label = self.labels[item]
#            if isinstance(label, unicode):
#                label = label.encode('utf-8')
                
            if not c.startswith('id') or mostrar_ids:
                columnas.append(remove_specials(self.cols[item]))
                labels.append(label)
        
        fichero_csv.add(labels)
        
        totales = self.init_totales()
        
        for dato in self.data:
            
            # totalizar
            if self.totales:
                for idx, c in enumerate(columnas):
                    if totales.has_key(c):
                        totales[c] = self.acumular(totales[c], dato[idx])
                        
            fila = []
            for col in columnas:
                fila.append(dato[col])                 
            
            fichero_csv.add(fila)
            
        # añadir línea con el total
        if self.totales:
            fila_totales = []
            for c in columnas:
                if totales.has_key(c):
                    fila_totales.append(self.float_fmt(totales[c]))
                    
                else:
                    fila_totales.append(None)
                    
            fichero_csv.add(fila_totales)
        
        return fichero_csv.read(encoding)
    
    def to_xls(self, title, filename, fmt=None):
        pl = PackageLoader('neptuno', 'templates')
        env = Environment(loader=pl)
        template = env.get_template('dstoxls.xml')
        
        xr = XLSReport(template)
        
        date_fmt = self.date_fmt.replace('%d', 'dd').\
            replace('%m', 'mm').\
            replace('%Y', 'yyyy')
            
        time_fmt = self.time_fmt.\
            replace('%H', 'HH').\
            replace('%M', 'MM').\
            replace('%S', 'SS')
        
        _fmt = dict(date=date_fmt, time=time_fmt)
        _fmt.update(fmt or {})
            
        params = dict(title=title, ds=self, fmt=_fmt)
        return xr.create(params, filename=filename)
    
    def order(self, claves):
        """
        IN
          claves <list> [(<col>, <signo>), ...]
            * <col>    Nombre de la columna
            * <signo>  +1 para ordenación ascendente
                       -1 para ordenación descendente
          
        OUT
          ???
        """

        def ordenar(ordenar_por):
            def _ordenar(function):
                def __ordenar(*args):
                    for o, signo in ordenar_por:
                        new_args = (args[0][o], args[1][o])
                        r = function(*new_args)
                        
                        if r != 0:
                            break
                    return r * signo
                
                return __ordenar
                        
            return _ordenar
        
        @ordenar(claves)
        def compara(x, y):
            if x < y:
                return -1
            
            elif x == y:
                return 0
            
            else:
                return 1
            
        self.data.sort(cmp=compara)
    
    @staticmethod
    def procesar_resultado(session, query, limit=None, pos=None, no_count=False, show_ids=False):
        """
        IN
          session  <sqlalchemy.orm.session.Session>
          query    <>
          limit    <int>
          pos      <int>
          no_count  <bool> = False
          show_ids  <bool> = False
          
        OUT
          <DataSet>
        """
        
        cols = []
        col_names = []
        for c in query.columns:
            if show_ids or not (c.name.startswith('id_') or c.name.startswith('_')):
                type_ = ''
                if isinstance(c.type, INTEGER) or isinstance(c.type, BIGINT):
                    type_ = 'int'
                    
                elif isinstance(c.type, NUMERIC):
                    type_ = 'float'
                    
                elif isinstance(c.type, DATE):
                    type_ = 'date'
                    
                elif isinstance(c.type, TIME):
                    type_ = 'time'
                    
                elif isinstance(c.type, DATETIME):
                    type_ = 'datetime'
                    
                elif isinstance(c.type, BOOLEAN):
                    type_ = 'bool'
                    
                #print c.type, type
                
                # avoid col name repeat
                name = remove_specials(c.name).lower()
                i = 1
                while name in col_names:
                    name = u'%s_%d' % (name, i)
                    i += 1
                    
                col_names.append(name)
                    
                cols.append((name, # name
                             c.name, # label
                             type_,))
                
        ds = DataSet(cols)
        if not no_count:
            ds.count = session.execute(query).rowcount
            
        else:
            ds.count = 0
        
        if limit == 0:
            limit = None
        
        for fila in session.execute(query.limit(limit).offset(pos)):
            row = []
            for c in query.columns:
                data = fila[c]
                if show_ids or not (c.name.startswith('id_') or c.name.startswith('_')):
                    if isinstance(data, str):
                        data = data.decode('utf-8')
                        
                    if isinstance(data, unicode):
                        data = data.replace(u'\u000B', ' ')
                    
                    row.append(data)

            ds.append(row)
            
            if no_count:
                ds.count += 1
                        
        return ds
    
    @staticmethod
    def query(dbs, query, cols, limit=None, pos=None, show_ids=False):
        """
        IN
          session  <sqlalchemy.orm.session.Session>
          query    <sqlalchemy.orm.query.Query>
          limit    <int>
          pos      <int>
          
        OUT
          <DataSet>
        """
        
        ds = DataSet(cols, limite=limit)
        ds.count = dbs.execute(query).rowcount
        
        if limit == 0:
            limit = None
        
        for row in dbs.execute(query.limit(limit).offset(pos)):
            ds.append(row.values())

        return ds


# -*- coding: utf-8 -*-

from email import Encoders
from email.header import Header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import os
import re
import smtplib

def send_mail(from_, to_, subject, message, server, login, password, files=[], 
              html='', cc=[], bcc=[], reply_to=None, charset='utf-8', ssl=False):
    """
    IN
      from_          (<e-mail address>, <name>)
      to_            [(<e-mail address>, <name>), ...]
      subject        <str>
      message        <str>
      server         <str> (smtp.domain.com, smtp.domain.com:<port>)
      login          <str>
      password       <str>
      files          [<str>, ...] (optional)
                     [(<StringIO>, <str>), ...] (optional)
      html           <str>        (optional)
      cc             [(<e-mail address>, <name>), ...]
      bcc            [(<e-mail address>, <name>), ...]
      reply_to       <str> => None
      charset        <str> => utf-8
      ssl            <bool> => False
      
      For example,
        send_mail(('pperez@gmail.com', 'Pepe Pérez'),
                  ['malopez@gmail.com', 'María López'), (mlopez@yahoo.com, 'María')],
                  'Hello, world!', 'How you doing?\n\nPepe',
                  'smtp.gmail.com', 
                  'pperez@gmail.com', 'ilovemarialopez')
                     
    """
    
    msg = MIMEText(message)
    msg.set_type('text/plain')
    msg.set_charset(charset)
    
    msg_root = None
    if html:
        msg_root = MIMEMultipart('related')
        msg_root.preamble = 'This is a multi-part message in MIME format.'
        
        if html:
            msg_alt = MIMEMultipart('alternative')
            
            msg_alt.attach(msg)
            
            msg_html = MIMEText(html)
            msg_html.set_type('text/html')
            msg_html.set_charset(charset)

            msg_alt.attach(msg_html)
            
            msg_root.attach(msg_alt)
            
        msg = msg_root

    files_obj = []        
    if files:
        if not msg_root:
            msg_root = MIMEMultipart()
            msg_root.attach(msg)
            msg = msg_root
            
        for f in files:
            part = MIMEBase('application', 'octet-stream')
            if isinstance(f, (str, unicode,)):
                f_name = os.path.basename(f)
                
                f = open(f, 'rb')
                files_obj.append(f)
                
            else:
                # tuple (<file-like>, <file name>,)
                f, f_name = f
                f.seek(0)
                
            part.set_payload(f.read())
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % f_name)
            msg.attach(part)
            
    msg['Subject'] = Header(subject, charset)
    msg['From'] = '%s <%s>' % (from_[1], from_[0])
    
    nombres = ['%s <%s>' % (dst[1], dst[0]) for dst in to_]
    msg['To'] = ';'.join(nombres)
    msg['Cc'] = ';'.join(['%s <%s>' % (dst[1], dst[0]) for dst in cc])
    
    # reply-to
    if reply_to:
        msg.add_header('Reply-to', reply_to)
    
    # identify the port in given host (smtp.domain.com:123)
    m_server = re.search(r'^([\w\-\.]+)(:\d+)?', server)
    port = m_server.group(2)
    if port:
        port = dict(port=int(port.replace(':', '')))

    else:
        port = {}
        
    SMTP_ = smtplib.SMTP
    if ssl:
        smtplib.SMTP_SSL
        
    s = SMTP_(m_server.group(1), **port)
    try:
        s.set_debuglevel(False)
        
        # authenticate to the server
        s.ehlo()
        
        # STARTTLS?
        if s.has_extn('STARTTLS'):
            s.starttls()
            s.ehlo() # re-identify ourselves over TLS connection
            
        # login
        s.login(login, password)
        
        emails_dst_to = [dst[0] for dst in to_]
        emails_dst_cc = [dst[0] for dst in cc]
        emails_dst_bcc = [dst[0] for dst in bcc]
        
        emails_dst = emails_dst_to + emails_dst_cc + emails_dst_bcc 
        
        # send e-mail
        result = s.sendmail(from_[0], emails_dst, msg.as_string())
        
        # close files (if there's any)
        for f in files_obj:
            f.close()
        
        return result
    
    finally:
        s.quit()
        
def normalizar_a_html(texto):
    return texto.\
        replace('\'', '&#39;').\
        replace('&', '&amp;').\
        replace('#', '&#35;').\
        replace('á', '&aacute;').\
        replace('Á', '&Aacute;').\
        replace('é', '&eacute;').\
        replace('É', '&Eacute;').\
        replace('í', '&iacute;').\
        replace('Í', '&Iacute;').\
        replace('ó', '&oacute;').\
        replace('Ó', '&Oacute;').\
        replace('ú', '&uacute;').\
        replace('Ú', '&Uacute;').\
        replace('ü', '&uuml;').\
        replace('Ü', '&Uuml;').\
        replace('ñ', '&ntilde;').\
        replace('Ñ', '&Ntilde;').\
        replace('<', '&lt;').\
        replace('>', '&gt;').\
        replace('¡', '&iexcl;').\
        replace('?', '&iquest;').\
        replace('"', '&quot;').\
        replace('%', '&#37;')
        



# -*- coding: utf-8 -*-

from setuptools import setup

setup(name='Neptuno2',
      version='1.1.0',
      author=u'León Domingo',
      author_email='leon.domingo@ender.es',
      description=('A little set of utilities that Ender has been using in their web projects'),
      #license=???,
      keywords='neptuno ender web',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
      ],
      url='http://www.ender.es',      
      packages=['neptuno',
                'neptuno.postgres',
                'neptuno.templates',
      ],
      package_data={
            'neptuno': ['default/*.*', 
                        'default/update/*.*',
                        'default/update/TODO',
                        'default/update/new-issue',
                        'default/update/ejemplo/*.*',                        
                        ],
            'neptuno.templates': ['*.xml'],
      },
      install_requires=[
        'psycopg2',
        'sqlalchemy',
        'lxml',
        'xlrd',
        'xlwt',
        'simplejson',
        'jinja2',
      ],
     )



# -*- coding: utf-8 -*-

import logging
import logging.handlers as loghandlers
import hashlib
import random
#import nucleo.config as config

LEVELS = dict(debug=logging.DEBUG,
              info=logging.INFO,
              warning=logging.WARNING,
              error=logging.ERROR,
              critical=logging.CRITICAL)

class NeptunoLogger(object):
    
    def __init__(self, nombre, fichero, nivel, tamanyo):
        self.nombre = nombre
        self.fichero = fichero
        self.nivel = nivel
        self.tamanyo = tamanyo
        
        if self.nivel:
            self.logger = logging.getLogger(self.nombre)
            self.logger.setLevel(LEVELS.get(self.nivel, logging.NOTSET))
    
            fh = loghandlers.RotatingFileHandler(self.fichero,
                                                 maxBytes=self.tamanyo, backupCount=5)
            self.logger.addHandler(fh)
            
            fmt = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            fh.setFormatter(fmt)
        
    def debug(self, msg):
        if self.nivel:
            self.logger.debug(msg)
            
    def info(self, msg):
        if self.nivel:
            self.logger.info(msg)
            
    def warning(self, msg):
        if self.nivel:
            self.logger.warning(msg)
            
    def error(self, msg):
        if self.nivel:
            self.logger.error(msg)
            
    def critical(self, msg):
        if self.nivel:
            self.logger.critical(msg)
    
    @staticmethod
    def get_logger(nombre):
        random.seed()
        m = hashlib.md5()
        m.update('%6.6d' % random.randint(0, 999999))
        
#        return NeptunoLogger('%s (%s)' % (nombre, m.hexdigest()), 
#                             config.LOG_FILE, config.LOG_LEVEL, config.LOG_SIZE)


# -*- coding: utf-8 -*-

from sqlalchemy.schema import Table, MetaData
from sqlalchemy.sql.expression import select

class ENoExisteRegistro(Exception):
    def __init__(self, id_reg):
        Exception.__init__(self)
        self.id_reg = id_reg
        
    def __repr__(self):
        return 'No existe el registro %d' % self.id_reg

def merge(nombre_tabla, id_destino, otros_ids, session):
    """
    Fusiona uno o más registros (otros_ids) en otro (id_destino), siendo todos ellos
    de la tabla 'nombre_tabla'.
    
    IN
      id_usuario   <int>: Identificador del usuario que se conecta
      nombre_tabla <str>: Nombre de la tabla a la que pertenecen los registros
      id_destino   <int>: Identificador del registro donde se guardará la información
      otros_ids    <list> [<int>, ...]: Lista de ids desde donde se obtendrá la información
    
    OUT
      un JSON de la forma:
      {
       'id_destino': 123,
       'id_origen': 123,
       'num_campos': 123
      }
    
    EXC
      ENoExisteRegistro: Cuando no existe alguno de los registros (origen o destino)
    """
    
    meta = MetaData(bind=session.bind, reflect=True)
    tabla = Table(nombre_tabla, meta, autoload=True)                
                        
    # comprobar que existe el registro "destino"
    alumno = session.execute(select([tabla], tabla.c.id == id_destino)).fetchone()
    if alumno is None:
        raise ENoExisteRegistro(id_destino)
    
    # comprobar que existen los "otros" registros
    for id_otro in otros_ids:
        
        otro = session.execute(select([tabla], tabla.c.id == id_otro)).fetchone()
        if otro is None:
            raise ENoExisteRegistro(id_otro)
            
    # fusionar
    resultado = {}
    resultado['id_destino'] = id_destino
    resultado['num_campos'] = 0         
    for id_otro in otros_ids:
        
        if id_otro == id_destino:
            continue
        
        resultado['id_origen'] = id_otro
        
        # obtener datos de los dos registros
        alumno = session.execute(select([tabla], tabla.c.id == id_destino)).fetchone()
        otro = session.execute(select([tabla], tabla.c.id == id_otro)).fetchone()
        
        # claves foráneas que hacen referencia a esta tabla
        for t in meta.sorted_tables:
            for fk in t.foreign_keys:
                if fk.references(tabla):
                    # actualizar los registros que apuntan a "origen" -> "destino"
                    qry_update = t.update(fk.parent == id_otro, values={fk.parent: id_destino})
                    session.execute(qry_update)
        
        session.commit()
        
        # cambiar datos de la tabla
        datos = {}
        for k in alumno.keys():
            if k != 'id' and k != 'busqueda':
                
                #print 'Tratando %s: %s - %s' % (k, alumno[k] or '<NULL>', otro[k] or '<NULL>')
                
#                if alumno[k] != None and otro[k] != None:
#                    if lista_atributos and str(k) in lista_atributos:
#                        merge('clientes', alumno[k], [otro[k]], conector=conector)
                
                if alumno[k] is None and otro[k] != None:
                    #print 'cambiando...'
                    datos[k] = otro[k]
                    resultado['num_campos'] += 1                   

        if datos != {}:
            # actualizar "destino"
            qry_update = tabla.update(tabla.c.id == id_destino, values=datos)
            session.execute(qry_update)

        # borrar "otro"
        qry_delete = tabla.delete(tabla.c.id == id_otro)
        session.execute(qry_delete)
        session.commit()                    
    
    return resultado